import sys
try:
    import __pypy__
    del __pypy__
    ispypy = True
except ImportError:
    ispypy = False


#- fake namespace for interactive lazy lookups -------------------------------
class InteractiveLazy(object):
    def __init__(self, hook_okay):
        self._hook_okay = hook_okay

    def __getattr__(self, attr):
        import cppyy, sys
        if attr == '__all__':
          # copy all exported items from cppyy itself
            for v in cppyy.__all__:
                self.__dict__[v] = getattr(cppyy, v)

          # add the lookup hook into cppyy.gbl if legal, or put it under 'g'
          # if not (PyPy and IPython for now)
            if self._hook_okay:
                caller = sys.modules[sys._getframe(1).f_globals['__name__']]
                cppyy._backend._set_cpp_lazy_lookup(caller.__dict__)
                return cppyy.__all__
            else:
                self.__dict__['g']   = cppyy.gbl
                self.__dict__['std'] = cppyy.gbl.std
                return ['g', 'std']+cppyy.__all__
        return getattr(cppyy, attr)

sys.modules['cppyy.interactive'] = InteractiveLazy(\
    not ispypy and not (hasattr(__builtins__, '__IPYTHON__') or 'IPython' in sys.modules))
del InteractiveLazy, ispypy



""" CPython-specific touch-ups
"""

from . import _stdcpp_fix
from cppyy_backend import loader

__all__ = [
    'gbl',
    'load_reflection_info',
    'addressof',
    'bind_object',
    'nullptr',
    '_backend',
    '_begin_capture_stderr',
    '_end_capture_stderr'
    ]

# first load the dependency libraries of the backend, then pull in the
# libcppyy extension module
c = loader.load_cpp_backend()
import libcppyy as _backend
_backend._cpp_backend = c

# explicitly expose APIs from libcppyy
import ctypes
_w = ctypes.CDLL(_backend.__file__, ctypes.RTLD_GLOBAL)


# some beautification for inspect (only on p2)
import sys
if sys.hexversion < 0x3000000:
  # TODO: this reliese on CPPOverload cooking up a func_code object, which atm
  # is simply not implemented for p3 :/

  # convince inspect that cppyy method proxies are possible drop-ins for python
  # methods and classes for pydoc
    import inspect

    inspect._old_isfunction = inspect.isfunction
    def isfunction(object):
        if type(object) == _backend.CPPOverload and not object.im_class:
            return True
        return inspect._old_isfunction(object)
    inspect.isfunction = isfunction

    inspect._old_ismethod = inspect.ismethod
    def ismethod(object):
        if type(object) == _backend.CPPOverload:
            return True
        return inspect._old_ismethod(object)
    inspect.ismethod = ismethod
    del isfunction, ismethod


### template support ---------------------------------------------------------
class Template(object):  # expected/used by ProxyWrappers.cxx in CPyCppyy
    stl_sequence_types   = ['std::vector', 'std::list', 'std::set', 'std::deque']
    stl_unrolled_types   = ['std::pair']
    stl_fixed_size_types = ['std::array']
    stl_mapping_types    = ['std::map']

    def __init__(self, name):
        self.__name__ = name

    def __repr__(self):
        return "<cppyy.Template '%s' object at %s>" % (self.__name__, hex(id(self)))

    def __getitem__(self, *args):
      # multi-argument to [] becomes a single tuple argument
        if args and type(args[0]) is tuple:
            args = args[0]

      # construct the type name from the types or their string representation
        newargs = [self.__name__]
        for arg in args:
            if type(arg) == str:
                arg = ','.join(map(lambda x: x.strip(), arg.split(',')))
            newargs.append(arg)
        pyclass = _backend.MakeCppTemplateClass(*newargs)

      # special case pythonization (builtin_map is not available from the C-API)
        if 'push_back' in pyclass.__dict__ and not '__iadd__' in pyclass.__dict__:
            if 'reserve' in pyclass.__dict__:
                def iadd(self, ll):
                    self.reserve(len(ll))
                    for x in ll: self.push_back(x)
                    return self
            else:
                def iadd(self, ll):
                    for x in ll: self.push_back(x)
                    return self
            pyclass.__iadd__ = iadd

        return pyclass

    def __call__(self, *args):
      # for C++17, we're required to derive the type when using initializer syntax
      # (i.e. a tuple or list); not sure how to do that in general, but below the
      # most common cases are covered
        if args:
            args0 = args[0]
            if args0 and (type(args0) is tuple or type(args0) is list):
                t = type(args0[0])
                if t is float: t = 'double'

                if self.__name__ in self.stl_sequence_types:
                    return self[t](*args)
                if self.__name__ in self.stl_fixed_size_types:
                    return self[t, len(args0)](*args)
                if self.__name__ in self.stl_unrolled_types:
                    return self[tuple(type(a) for a in args0)](*args0)

            if args0 and type(args0) is dict:
                if self.__name__ in self.stl_mapping_types:
                    try:
                        pair = args0.items().__iter__().__next__()
                    except AttributeError:
                        pair = args0.items()[0]
                    t1 = type(pair[0])
                    if t1 is float: t1 = 'double'
                    t2 = type(pair[1])
                    if t2 is float: t2 = 'double'
                    return self[t1, t2](*args)

                return self.__getitem__(*(type(a) for a in args0))(*args)

      # old 'metaclass-style' template instantiations
        return self.__getitem__(*args)

_backend.Template = Template


#- :: and std:: namespaces ---------------------------------------------------
gbl = _backend.CreateScopeProxy('')
gbl.__class__.__repr__ = lambda cls : '<namespace cppyy.gbl at 0x%x>' % id(cls)
gbl.std =  _backend.CreateScopeProxy('std')
# for move, we want our "pythonized" one, not the C++ template
gbl.std.move  = _backend.move


#- add to the dynamic path as needed -----------------------------------------
import os
def add_default_paths():
    gSystem = gbl.gSystem
    if os.getenv('CONDA_PREFIX'):
      # MacOS, Linux
        lib_path = os.path.join(os.getenv('CONDA_PREFIX'), 'lib')
        if os.path.exists(lib_path): gSystem.AddDynamicPath(lib_path)

      # Windows
        lib_path = os.path.join(os.getenv('CONDA_PREFIX'), 'Library', 'lib')
        if os.path.exists(lib_path): gSystem.AddDynamicPath(lib_path)

  # assuming that we are in PREFIX/lib/python/site-packages/cppyy, add PREFIX/lib to the search path
    lib_path = os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir, os.path.pardir))
    if os.path.exists(lib_path): gSystem.AddDynamicPath(lib_path)

    try:
        with open('/etc/ld.so.conf') as ldconf:
            for line in ldconf:
                f = line.strip()
                if (os.path.exists(f)):
                    gSystem.AddDynamicPath(f)
    except IOError:
        pass
add_default_paths()
del add_default_paths


#- exports -------------------------------------------------------------------
addressof     = _backend.addressof
bind_object   = _backend.bind_object
nullptr       = _backend.nullptr

def load_reflection_info(name):
    sc = gbl.gSystem.Load(name)
    if sc == -1:
        raise RuntimeError("Unable to load reflection library "+name)

def _begin_capture_stderr():
    _backend._begin_capture_stderr()

def _end_capture_stderr():
    err = _backend._end_capture_stderr()
    if err:
        return "\n%s" % err
    return ""



""" Externally provided types: get looked up if all else fails, e.g.
    for typedef-ed C++ builtin types.
"""

import sys

def _create_mapper(cls, extra_dct=None):
    def mapper(name, scope):
        if scope:
            cppname = scope+'::'+name
            modname = 'cppyy.gbl.'+scope
        else:
            cppname = name
            modname = 'cppyy.gbl'
        dct = {'__cpp_name__' : cppname, '__module__' : modname}
        if extra_dct: dct.update(extra_dct)
        return type(name, (cls,), dct)
    return mapper

# from six.py ---
# Copyright (c) 2010-2017 Benjamin Peterson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""
    # This requires a bit of explanation: the basic idea is to make a dummy
    # metaclass for one level of class instantiation that replaces itself with
    # the actual metaclass.
    class metaclass(type):

        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)

        @classmethod
        def __prepare__(cls, name, this_bases):
            return meta.__prepare__(name, bases)
    return type.__new__(metaclass, 'temporary_class', (), {})
# --- end from six.py

class _BoolMeta(type):
    def __call__(self, val = bool()):
        if val: return True
        else: return False

class _Bool(with_metaclass(_BoolMeta, object)):
    pass


def initialize(backend):
    if not hasattr(backend, 'type_map'):
        return

    tm = backend.type_map

    # boolean type (builtin type bool can nog be subclassed)
    tm['bool'] = _create_mapper(_Bool)

    # char types
    str_tm = _create_mapper(str)
    for tp in ['char', 'unsigned char', 'signed char']:
        tm[tp] = str_tm
    if sys.hexversion < 0x3000000:
        tm['wchar_t'] = _create_mapper(unicode)
    else:
        tm['wchar_t'] = str_tm

    # integer types
    int_tm = _create_mapper(int)
    for tp in ['short', 'unsigned short', 'int']:
        tm[tp] = int_tm

    if sys.hexversion < 0x3000000:
        long_tm = _create_mapper(long)
    else:
        long_tm = tm['int']
    for tp in ['unsigned int', 'long', 'unsigned long', 'long long', 'unsigned long long']:
        tm[tp] = long_tm

    # floating point types
    float_tm = _create_mapper(float)
    for tp in ['float', 'double', 'long double']:
        tm[tp] = float_tm

    # void*
    import ctypes
    def voidp_init(self, arg=0):
        import cppyy, ctypes
        if arg == cppyy.nullptr: arg = 0
        ctypes.c_void_p.__init__(self, arg)
    tm['void*'] = _create_mapper(ctypes.c_void_p, {'__init__' : voidp_init})



""" C++ proxy types.
"""

import cppyy

bck = cppyy._backend
Instance      = bck.CPPInstance

try:
    import __pypy__
    __all__ = [
        'Instance'
    ]

except ImportError:
    __all__ = [
        'DataMember',
        'Instance',
        'Function',
        'Method'
        'Scope',
        'InstanceArray',
        'LowLevelView',
        'Template'
    ]

    DataMember      = bck.CPPDataMember
    Function        = bck.CPPOverload
    Method          = bck.CPPOverload
    Scope           = bck.CPPScope
    InstanceArray   = bck.InstanceArray
    LowLevelView    = bck.LowLevelView
    Template        = bck.TemplateProxy

del bck



# PyInstaller hooks to declare the "data files" (libraries, headers, etc.) of
# cppyy-backend. Placed here rather then in cppyy-backend to guarantee that any
# packaging of top-level "cppyy" picks up the backend as well.
#
# See also setup.cfg.

__all__ = ['data']


def _backend_files():
    import cppyy_backend, glob, os

    all_files = glob.glob(os.path.join(
        os.path.dirname(cppyy_backend.__file__), '*'))

    def datafile(path):
        return path, os.path.join('cppyy_backend', os.path.basename(path))

    return [datafile(filename) for filename in all_files if os.path.isdir(filename)]

def _api_files():
    import cppyy, os

    paths = str(cppyy.gbl.gInterpreter.GetIncludePath()).split('-I')
    for p in paths:
        if not p: continue

        apipath = os.path.join(p.strip()[1:-1], 'CPyCppyy')
        if os.path.exists(apipath):
            return [(apipath, os.path.join('include', 'CPyCppyy'))]

    return []

datas = _backend_files()+_api_files()



""" Low-level utilities, to be used for "emergencies only".
"""

import cppyy

try:
    import __pypy__
    del __pypy__
    ispypy = True
except ImportError:
    ispypy = False

__all__ = [
    'cast',
    'static_cast',
    'reinterpret_cast',
    'dynamic_cast',
    'malloc',
    'free',
    'array_new',
    'array_delete',
    'signals_as_exception',
    'set_signals_as_exception'
    'FatalError'
    'BusError',
    'SegmentationViolation',
    'IllegalInstruction',
    'AbortSignal',
    ]


# import low-level python converters
for _name in ['addressof', 'as_cobject', 'as_capsule', 'as_ctypes']:
    try:
        exec('%s = cppyy._backend.%s' % (_name, _name))
        __all__.append(_name)
    except AttributeError:
        pass
del _name


# create low-level helpers
cppyy.cppdef("""namespace __cppyy_internal {
// type casting
    template<typename T, typename U>
    T cppyy_cast(U val) { return (T)val; }

    template<typename T, typename U>
    T cppyy_static_cast(U val) { return static_cast<T>(val); }

    template<typename T, typename U>
    T cppyy_reinterpret_cast(U val) { return reinterpret_cast<T>(val); }

    template<typename T, typename S>
    T* cppyy_dynamic_cast(S* obj) { return dynamic_cast<T*>(obj); }

// memory allocation/free-ing
    template<typename T>
    T* cppyy_malloc(size_t count=1) { return (T*)malloc(sizeof(T*)*count); }

    template<typename T>
    T* cppyy_array_new(size_t count) { return new T[count]; }

    template<typename T>
    void cppyy_array_delete(T* ptr) { delete[] ptr; }
}""")


# helper for sizing arrays
class ArraySizer(object):
    def __init__(self, func):
        self.func = func
    def __getitem__(self, t):
        self.array_type = t
        return self
    def __call__(self, size, managed=False):
        res = self.func[self.array_type](size)
        res.reshape((size,))
        if managed: res.__python_owns__ = True
        return res

class CArraySizer(ArraySizer):
    def __call__(self, size, managed=False):
        res = ArraySizer.__call__(self, size, managed)
        res.__cpp_array__ = False
        return res


# import casting helpers
cast             = cppyy.gbl.__cppyy_internal.cppyy_cast
static_cast      = cppyy.gbl.__cppyy_internal.cppyy_static_cast
reinterpret_cast = cppyy.gbl.__cppyy_internal.cppyy_reinterpret_cast
dynamic_cast     = cppyy.gbl.__cppyy_internal.cppyy_dynamic_cast

# import memory allocation/free-ing helpers
malloc           = CArraySizer(cppyy.gbl.__cppyy_internal.cppyy_malloc)
free             = cppyy.gbl.free      # for symmetry
array_new        = ArraySizer(cppyy.gbl.__cppyy_internal.cppyy_array_new)
array_delete     = cppyy.gbl.__cppyy_internal.cppyy_array_delete

# signals as exceptions
if not ispypy:
    FatalError            = cppyy._backend.FatalError
    BusError              = cppyy._backend.BusError
    SegmentationViolation = cppyy._backend.SegmentationViolation
    IllegalInstruction    = cppyy._backend.IllegalInstruction
    AbortSignal           = cppyy._backend.AbortSignal

    class signals_as_exception:
        def __enter__(self):
            cppyy._backend.SetGlobalSignalPolicy(1)

        def __exit__(self, type, value, traceback):
            cppyy._backend.SetGlobalSignalPolicy(0)

    set_signals_as_exception = cppyy._backend.SetGlobalSignalPolicy

else:
    class FatalError(Exception):
        pass
    class BusError(FatalError):
        pass
    class SegmentationViolation(FatalError):
        pass
    class IllegalInstruction(FatalError):
        pass
    class AbortSignal(FatalError):
        pass

    class signals_as_exception:
        def __enter__(self):
            pass   # not yet implemented

        def __exit__(self, type, value, traceback):
            pass   # not yet implemented

    def set_signals_as_exception(seton):
        return False

del ispypy




""" PyPy-specific touch-ups
"""

from . import _stdcpp_fix

import os, sys
from cppyy_backend import loader

__all__ = [
    'gbl',
    'addressof',
    'bind_object',
    'nullptr',
    ]

# first load the dependency libraries of the backend, then
# pull in the built-in low-level cppyy
try:
    c = loader.load_cpp_backend()
except RuntimeError:
    import sysconfig
    os.environ['CPPYY_BACKEND_LIBRARY'] = "libcppyy_backend"+sysconfig.get_config_var("SO")
    c = loader.load_cpp_backend()
os.environ['CPPYY_BACKEND_LIBRARY'] = c._name

# some older versions can be fixed up through a compatibility
# module on the python side; load it, if available
try:
    import cppyy_compat
except ImportError:
    pass

import _cppyy as _backend     # built-in module
try:
    _backend._post_import_startup()
except AttributeError:
    pass
_backend._cpp_backend = c

def fixup_legacy():
    version = sys.pypy_version_info
    if version[0] < 7 or (version[0] == 7 and version[1] <= 3 and version[2] <= 3):
        _backend.gbl.CppyyLegacy = _backend.gbl
fixup_legacy()
del fixup_legacy


#- exports -------------------------------------------------------------------
import sys
_thismodule = sys.modules[__name__]
for name in __all__:
    try:
        setattr(_thismodule, name, getattr(_backend, name))
    except AttributeError:
        pass
del name, sys
nullptr = _backend.nullptr

def load_reflection_info(name):
    sc = _backend.gbl.gSystem.Load(name)
    if sc == -1:
        raise RuntimeError("Unable to load reflection library "+name)

def _begin_capture_stderr():
    pass

def _end_capture_stderr():
    return ""

# add other exports to all
__all__.append('load_reflection_info')
__all__.append('_backend')
__all__.append('_begin_capture_stderr')
__all__.append('_end_capture_stderr')



"""Dynamic C++ bindings generator.

This module provides dynamic bindings to C++ through Cling, the LLVM-based C++
interpreter, allowing interactive mixing of Python and C++. Example:

    >>> import cppyy
    >>> cppyy.cppdef(\"\"\"
    ... class MyClass {
    ... public:
    ...     MyClass(int i) : m_data(i) {}
    ...     int m_data;
    ... };\"\"\")
    True
    >>> from cppyy.gbl import MyClass
    >>> m = MyClass(42)
    >>> cppyy.cppdef(\"\"\"
    ... void say_hello(MyClass* m) {
    ...     std::cout << "Hello, the number is: " << m->m_data << std::endl;
    ... }\"\"\")
    True
    >>> MyClass.say_hello = cppyy.gbl.say_hello
    >>> m.say_hello()
    Hello, the number is: 42
    >>> m.m_data = 13
    >>> m.say_hello()
    Hello, the number is: 13
    >>>

For full documentation, see:
   https://cppyy.readthedocs.io/

"""

__author__ = 'Wim Lavrijsen <WLavrijsen@lbl.gov>'

__all__ = [
    'cppdef',                 # declare C++ source to Cling
    'cppexec',                # execute a C++ statement
    'include',                # load and jit a header file
    'c_include',              # load and jit a C header file
    'load_library',           # load a shared library
    'nullptr',                # unique pointer representing NULL
    'sizeof',                 # size of a C++ type
    'typeid',                 # typeid of a C++ type
    'multi',                  # helper for multiple inheritance
    'add_include_path',       # add a path to search for headers
    'add_library_path',       # add a path to search for headers
    'add_autoload_map',       # explicitly include an autoload map
    'set_debug',              # enable/disable debug output
    ]

from ._version import __version__

import ctypes, os, sys, sysconfig, warnings

if not 'CLING_STANDARD_PCH' in os.environ:
    local_pch = os.path.join(os.path.dirname(__file__), 'allDict.cxx.pch')
    if os.path.exists(local_pch):
        os.putenv('CLING_STANDARD_PCH', local_pch)
        os.environ['CLING_STANDARD_PCH'] = local_pch

try:
    import __pypy__
    del __pypy__
    ispypy = True
except ImportError:
    ispypy = False

# import separately instead of in the above try/except block for easier to
# understand tracebacks
if ispypy:
    from ._pypy_cppyy import *
else:
    from ._cpython_cppyy import *


#- allow importing from gbl --------------------------------------------------
sys.modules['cppyy.gbl'] = gbl
sys.modules['cppyy.gbl.std'] = gbl.std


#- enable auto-loading -------------------------------------------------------
try:    gbl.gInterpreter.EnableAutoLoading()
except: pass


#- external typemap ----------------------------------------------------------
from . import _typemap
_typemap.initialize(_backend)


#- pythonization factories ---------------------------------------------------
from . import _pythonization as py
py._set_backend(_backend)

def _standard_pythonizations(pyclass, name):
  # pythonization of tuple; TODO: placed here for convenience, but a custom case
  # for tuples on each platform can be made much more performant ...
    if name.find('tuple<', 0, 6) == 0:
        import cppyy
        pyclass._tuple_len = cppyy.gbl.std.tuple_size(pyclass).value
        def tuple_len(self):
            return self.__class__._tuple_len
        pyclass.__len__ = tuple_len
        def tuple_getitem(self, idx, get=cppyy.gbl.std.get):
            if idx < self.__class__._tuple_len:
                res = get[idx](self)
                try:
                    res.__life_line = self
                except Exception:
                    pass
                return res
            raise IndexError(idx)
        pyclass.__getitem__ = tuple_getitem

  # pythoniztion of std::string; placed here because it's simpler to write the
  # custom "npos" object (to allow easy result checking of find/rfind) in Python
    elif pyclass.__cpp_name__ == "std::string":
        class NPOS(0x3000000 <= sys.hexversion and int or long):
            def __eq__(self, other):
                return other == -1 or  int(self) == other
            def __ne__(self, other):
                return other != -1 and int(self) != other
        del pyclass.__class__.npos          # drop b/c is const data
        pyclass.npos = NPOS(pyclass.npos)

    return True

if not ispypy:
    py.add_pythonization(_standard_pythonizations, "std")
# TODO: PyPy still has the old-style pythonizations, which require the full
# class name (not possible for std::tuple ...)

# std::make_shared/unique create needless templates: rely on Python's introspection
# instead. This also allows Python derived classes to be handled correctly.
class py_make_smartptr(object):
    __slots__ = ['cls', 'ptrcls']
    def __init__(self, cls, ptrcls):
        self.cls    = cls
        self.ptrcls = ptrcls
    def __call__(self, *args):
        if len(args) == 1 and type(args[0]) == self.cls:
            obj = args[0]
        else:
            obj = self.cls(*args)
        return self.ptrcls[self.cls](obj)   # C++ takes ownership

class make_smartptr(object):
    __slots__ = ['ptrcls', 'maker']
    def __init__(self, ptrcls, maker):
        self.ptrcls = ptrcls
        self.maker  = maker
    def __call__(self, ptr):
        return py_make_smartptr(type(ptr), self.ptrcls)(ptr)
    def __getitem__(self, cls):
        try:
            if not cls.__module__ == int.__module__:
                return py_make_smartptr(cls, self.ptrcls)
        except AttributeError:
            pass
        if type(cls) == str and not cls in ('int', 'float'):
            return py_make_smartptr(getattr(gbl, cls), self.ptrcls)
        return self.maker[cls]

gbl.std.make_shared = make_smartptr(gbl.std.shared_ptr, gbl.std.make_shared)
gbl.std.make_unique = make_smartptr(gbl.std.unique_ptr, gbl.std.make_unique)
del make_smartptr


#--- interface to Cling ------------------------------------------------------
class _stderr_capture(object):
    def __init__(self):
       self._capture = not gbl.CppyyLegacy.gDebug and True or False
       self.err = ""

    def __enter__(self):
        if self._capture:
           _begin_capture_stderr()
        return self

    def __exit__(self, tp, val, trace):
        if self._capture:
            self.err = _end_capture_stderr()

def cppdef(src):
    """Declare C++ source <src> to Cling."""
    with _stderr_capture() as err:
        errcode = gbl.gInterpreter.Declare(src)
    if not errcode:
        raise SyntaxError('Failed to parse the given C++ code%s' % err.err)
    return True

def cppexec(stmt):
    """Execute C++ statement <stmt> in Cling's global scope."""
    if stmt and stmt[-1] != ';':
        stmt += ';'

  # capture stderr, but note that ProcessLine could legitimately be writing to
  # std::cerr, in which case the captured output needs to be printed as normal
    with _stderr_capture() as err:
        errcode = ctypes.c_int(0)
        gbl.gInterpreter.ProcessLine(stmt, ctypes.pointer(errcode))

    if errcode.value:
        raise SyntaxError('Failed to parse the given C++ code%s' % err.err)
    elif err.err and err.err[1:] != '\n':
        sys.stderr.write(err.err[1:])

    return True

def load_library(name):
    """Explicitly load a shared library."""
    with _stderr_capture() as err:
        gSystem = gbl.gSystem
        if name[:3] != 'lib':
            if not gSystem.FindDynamicLibrary(gbl.CppyyLegacy.TString(name), True) and\
                   gSystem.FindDynamicLibrary(gbl.CppyyLegacy.TString('lib'+name), True):
                name = 'lib'+name
        sc = gSystem.Load(name)
    if sc == -1:
        raise RuntimeError('Unable to load library "%s"%s' % (name, err.err))
    return True

def include(header):
    """Load (and JIT) header file <header> into Cling."""
    with _stderr_capture() as err:
        errcode = gbl.gInterpreter.Declare('#include "%s"' % header)
    if not errcode:
        raise ImportError('Failed to load header file "%s"%s' % (header, err.err))
    return True

def c_include(header):
    """Load (and JIT) header file <header> into Cling."""
    with _stderr_capture() as err:
        errcode = gbl.gInterpreter.Declare("""extern "C" {
#include "%s"
}""" % header)
    if not errcode:
        raise ImportError('Failed to load header file "%s"%s' % (header, err.err))
    return True

def add_include_path(path):
    """Add a path to the include paths available to Cling."""
    if not os.path.isdir(path):
        raise OSError('No such directory: %s' % path)
    gbl.gInterpreter.AddIncludePath(path)

def add_library_path(path):
    """Add a path to the library search paths available to Cling."""
    if not os.path.isdir(path):
        raise OSError('No such directory: %s' % path)
    gbl.gSystem.AddDynamicPath(path)

# add access to Python C-API headers
apipath = sysconfig.get_path('include', 'posix_prefix' if os.name == 'posix' else os.name)
if os.path.exists(apipath):
    add_include_path(apipath)
elif ispypy:
  # possibly structured without 'pythonx.y' in path
    apipath = os.path.dirname(apipath)
    if os.path.exists(apipath) and os.path.exists(os.path.join(apipath, 'Python.h')):
        add_include_path(apipath)

# add access to extra headers for dispatcher (CPyCppyy only (?))
if not ispypy:
    try:
        apipath_extra = os.environ['CPPYY_API_PATH']
        if os.path.basename(apipath_extra) == 'CPyCppyy':
            apipath_extra = os.path.dirname(apipath_extra)
    except KeyError:
        apipath_extra = None

    if apipath_extra is None:
        try:
            import pkg_resources as pr

            d = pr.get_distribution('CPyCppyy')
            for line in d.get_metadata_lines('RECORD'):
                if 'API.h' in line:
                    part = line[0:line.find(',')]

            ape = os.path.join(d.location, part)
            if os.path.exists(ape):
                apipath_extra = os.path.dirname(os.path.dirname(ape))

            del part, d, pr
        except Exception:
            pass

    if apipath_extra is None:
        ldversion = sysconfig.get_config_var('LDVERSION')
        if not ldversion: ldversion = sys.version[:3]

        apipath_extra = os.path.join(os.path.dirname(apipath), 'site', 'python'+ldversion)
        if not os.path.exists(os.path.join(apipath_extra, 'CPyCppyy')):
            import glob, libcppyy
            ape = os.path.dirname(libcppyy.__file__)
          # a "normal" structure finds the include directory up to 3 levels up,
          # ie. dropping lib/pythonx.y[md]/site-packages
            for i in range(3):
                if os.path.exists(os.path.join(ape, 'include')):
                    break
                ape = os.path.dirname(ape)

            ape = os.path.join(ape, 'include')
            if os.path.exists(os.path.join(ape, 'CPyCppyy')):
                apipath_extra = ape
            else:
              # add back pythonx.y or site/pythonx.y if present
                for p in glob.glob(os.path.join(ape, 'python'+sys.version[:3]+'*'))+\
                         glob.glob(os.path.join(ape, '*', 'python'+sys.version[:3]+'*')):
                    if os.path.exists(os.path.join(p, 'CPyCppyy')):
                        apipath_extra = p
                        break

    if apipath_extra.lower() != 'none':
        if not os.path.exists(os.path.join(apipath_extra, 'CPyCppyy')):
            warnings.warn("CPyCppyy API not found (tried: %s); set CPPYY_API_PATH envar to the 'CPyCppyy' API directory to fix" % apipath_extra)
        else:
            add_include_path(apipath_extra)

    del apipath_extra

if os.getenv('CONDA_PREFIX'):
  # MacOS, Linux
    include_path = os.path.join(os.getenv('CONDA_PREFIX'), 'include')
    if os.path.exists(include_path): add_include_path(include_path)

  # Windows
    include_path = os.path.join(os.getenv('CONDA_PREFIX'), 'Library', 'include')
    if os.path.exists(include_path): add_include_path(include_path)

# assuming that we are in PREFIX/lib/python/site-packages/cppyy, add PREFIX/include to the search path
include_path = os.path.abspath(os.path.join(os.path.dirname(__file__), *(4*[os.path.pardir]+['include'])))
if os.path.exists(include_path): add_include_path(include_path)

del include_path, apipath, ispypy

def add_autoload_map(fname):
    """Add the entries from a autoload (.rootmap) file to Cling."""
    if not os.path.isfile(fname):
        raise OSError("no such file: %s" % fname)
    gbl.gInterpreter.LoadLibraryMap(fname)

def set_debug(enable=True):
    """Enable/disable debug output."""
    if enable:
        gbl.CppyyLegacy.gDebug = 10
    else:
        gbl.CppyyLegacy.gDebug =  0

def _get_name(tt):
    if type(tt) == str:
        return tt
    try:
        ttname = tt.__cpp_name__
    except AttributeError:
        ttname = tt.__name__
    return ttname

_sizes = {}
def sizeof(tt):
    """Returns the storage size (in chars) of C++ type <tt>."""
    if not isinstance(tt, type) and not type(tt) == str:
        tt = type(tt)
    try:
        return _sizes[tt]
    except KeyError:
        sz = gbl.gInterpreter.ProcessLine("sizeof(%s);" % (_get_name(tt),))
        _sizes[tt] = sz
        return sz

_typeids = {}
def typeid(tt):
    """Returns the C++ runtime type information for type <tt>."""
    if not isinstance(tt, type):
        tt = type(tt)
    try:
        return _typeids[tt]
    except KeyError:
        tidname = 'typeid_'+str(len(_typeids))
        gbl.gInterpreter.ProcessLine(
            "namespace _cppyy_internal { auto* %s = &typeid(%s); }" %\
            (tidname, _get_name(tt),))
        tid = getattr(gbl._cppyy_internal, tidname)
        _typeids[tt] = tid
        return tid

def multi(*bases):      # after six, see also _typemap.py
    """Resolve metaclasses for multiple inheritance."""
  # contruct a "no conflict" meta class; the '_meta' is needed by convention
    nc_meta = type.__new__(type, 'cppyy_nc_meta', tuple(type(b) for b in bases if type(b) is not type), {})
    class faux_meta(type):
        def __new__(cls, name, this_bases, d):
            return nc_meta(name, bases, d)
    return type.__new__(faux_meta, 'faux_meta', (), {})


#- workaround (TODO: may not be needed with Clang9) --------------------------
if 'win32' in sys.platform:
    cppdef("""template<>
    std::basic_ostream<char, std::char_traits<char>>& __cdecl std::endl<char, std::char_traits<char>>(
        std::basic_ostream<char, std::char_traits<char>>&);""")



# monkey patch to be able to select a specific backend based on PyPy's version,
# which is not possible in the pyproject.toml file as there is currently no
# marker for it (this may change, after which this file can be removed)

try:
  # _BACKEND is the primary, __legacy__ the backwards compatible backend
    from setuptools.build_meta import _BACKEND
    main = _BACKEND
except (NameError, ImportError):
  # fallback as the name __legacy__ is actually documented (and part of __all__)
    main = __legacy__

# the following ensures proper build/installation order, after which the normal
# install through setup.py picks up their wheels from the cache (TODO: note the
# duplication here with setup.py; find a better way)
_get_requires_for_build_wheel = main.get_requires_for_build_wheel
def get_requires_for_build_wheel(*args, **kwds):
    try:
        import __pypy__, sys
        version = sys.pypy_version_info
        requirements = ['cppyy-cling==6.25.0']
        if version[0] == 5:
            if version[1] <= 9:
                requirements = ['cppyy-cling<6.12']
            elif version[1] <= 10:
                requirements = ['cppyy-cling<=6.15']
        elif version[0] == 6:
            if version[1] <= 0:
                requirements = ['cppyy-cling<=6.15']
        elif version[0] == 7:
            if version[1] <= 3 and version[2] <= 3:
                requirements = ['cppyy-cling<=6.18.2.3']
    except ImportError:
        # CPython
        requirements = ['cppyy-backend==1.14.5', 'cppyy-cling==6.25.0']

    return requirements + _get_requires_for_build_wheel(*args, **kwds)

main.get_requires_for_build_wheel = get_requires_for_build_wheel
